"""Example demonstrating how to use the ContextGroundingVectorStore class with LangChain."""

import argparse
import asyncio
from pprint import pprint
from typing import Any

from dotenv import find_dotenv, load_dotenv
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_core.vectorstores import VectorStore
from uipath_langchain.chat.models import UiPathAzureChatOpenAI
from uipath_langchain.vectorstores.context_grounding_vectorstore import (
    ContextGroundingVectorStore,
)


def create_retrieval_chain(vectorstore: VectorStore, model: BaseChatModel, k: int = 3):
    """Create a retrieval chain using a vector store.

    Args:
        vectorstore: Vector store to use for the chain
        model: LangChain language model to use for the chain

    Returns:
        A retrieval chain ready to answer questions
    """
    # Create a retriever from the vector store
    retriever = vectorstore.as_retriever(
        search_kwargs={"k": k},
    )

    # Create a prompt template
    template = """Answer the question based on the following context:
    {context}
    Question: {question}
    """
    prompt = ChatPromptTemplate.from_template(template)

    # Create the retrieval chain
    chain = (
        {"context": retriever, "question": RunnablePassthrough()}
        | prompt
        | model
        | StrOutputParser()
    )

    # Return a function that will run the chain and include source documents
    def retrieval_chain(query: str) -> dict[str, Any]:
        # Get documents separately to include them in the result
        docs = retriever.invoke(query)
        # Run the chain
        answer = chain.invoke(query)
        # Return combined result
        return {"result": answer, "source_documents": docs}

    return retrieval_chain


async def main(index_name: str, query: str, k: int = 3):
    load_dotenv(find_dotenv())

    """Run a simple example of ContextGroundingVectorStore."""
    vectorstore = ContextGroundingVectorStore(
        index_name=index_name,
    )

    # Example query
    query = "What is the ECCN for a laptop?"

    # Perform semantic searches with distance scores
    docs_with_scores = await vectorstore.asimilarity_search_with_score(query=query, k=5)
    print("==== Docs with distance scores ====")
    pprint(
        [
            {"page_content": doc.page_content, "distance_score": distance_score}
            for doc, distance_score in docs_with_scores
        ]
    )

    # Perform a similarity search with relevance scores
    docs_with_relevance_scores = (
        await vectorstore.asimilarity_search_with_relevance_scores(query=query, k=5)
    )
    print("==== Docs with relevance scores ====")
    pprint(
        [
            {"page_content": doc.page_content, "relevance_score": relevance_score}
            for doc, relevance_score in docs_with_relevance_scores
        ]
    )

    # Run a retrieval chain
    model = UiPathAzureChatOpenAI(
        model="gpt-4o-2024-08-06",
        max_retries=3,
    )

    retrieval_chain = create_retrieval_chain(
        vectorstore=vectorstore,
        model=model,
    )

    # Run a retrieval chain
    result = retrieval_chain(query)
    print("==== Retrieval chain result ====")
    print(f"Query: {query}")
    print(f"Answer: {result['result']}")
    print("\nSource Documents:")
    for i, doc in enumerate(result["source_documents"]):
        print(f"\nDocument {i + 1}:")
        print(f"Content: {doc.page_content[:100]}...")
        print(
            f"Source: {doc.metadata.get('source', 'N/A')}, Page Number: {doc.metadata.get('page_number', '0')}"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--index_name", type=str, default="ECCN", help="The name of the index to use"
    )
    parser.add_argument(
        "--query",
        type=str,
        default="What is the ECCN for a laptop?",
        help="The query for which documents will be retrieved",
    )
    parser.add_argument(
        "--k", type=int, default=3, help="The number of documents to retrieve"
    )
    args = parser.parse_args()
    asyncio.run(main(args.index_name, args.query, args.k))



from fastapi import FastAPI, Request, Depends, HTTPException, BackgroundTasks
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy.orm import Session
from typing import List, Optional
import httpx
import uuid
from datetime import datetime
from uipath import UiPath

# Database models and setup
from database import SessionLocal, engine
import models
import schemas

sdk = UiPath()

# Create database tables
models.Base.metadata.create_all(bind=engine)

app = FastAPI(title="Open Deep Research")

# Enable CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

templates = Jinja2Templates(directory="templates")

def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()

@app.get("/", response_class=HTMLResponse)
async def index(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})

# API endpoint to submit a new job
@app.post("/api/submit-job", response_model=schemas.JobResponse)
async def submit_job(job_request: schemas.JobRequest, background_tasks: BackgroundTasks, db: Session = Depends(get_db)):
    # Create a new job record in the database
    job_id = str(uuid.uuid4())
    new_job = models.Job(
        id=job_id,
        topic=job_request.topic,
        status="creating",
        created_at=datetime.now()
    )
    db.add(new_job)
    db.commit()
    db.refresh(new_job)


    # Submit the job to the external service in the background
    background_tasks.add_task(
        submit_orchestrator_job,
        job_id=job_id,
        topic=job_request.topic,
        db=db
    )

    return {"job_id": job_id, "status": "creating", "message": "Job submitted successfully"}

# Background task to submit job to Orchestrator
async def submit_orchestrator_job(job_id: str, topic: str, db: Session):
    try:
        agent_job = await sdk.processes.invoke_async(
            name="open_deep_research",
            input_arguments={"topic": topic}
        )
        print(agent_job)
        # Update job with Orchestrator job key if provided
        job = db.query(models.Job).filter(models.Job.id == job_id).first()
        if job and agent_job:
            job.agent_job_key = agent_job.Key
            job.agent_job_id = agent_job.Id
            job.status = "pending"
            db.commit()
        print(job.agent_job_id)

    except httpx.HTTPError as e:
        # If Orchestrator request fails, update the job status
        job = db.query(models.Job).filter(models.Job.id == job_id).first()
        if job:
            job.status = "faulted"
            job.error_message = str(e)
            db.commit()

# Webhook endpoint for Orchestrator job events notifications
@app.post("/webhook/job-complete")
async def job_complete_webhook(request: Request, db: Session = Depends(get_db)):
    webhook_data = await request.json()
    agent_job = webhook_data.get('Job')
    agent_job_id = agent_job.get('Id')
    #agent_job_key = agent_job.get('Key')

    print(webhook_data)
    print(agent_job_id)
    if not agent_job_id:
        raise HTTPException(status_code=400, detail="Missing Job.Key in webhook data")

    # Retrieve the job from the database
    job = db.query(models.Job).filter(models.Job.agent_job_id == agent_job_id).first()
    if not job:
        return {"status": "faulted", "message": "Job not found"}

    event_type = webhook_data.get('Type')

    print(event_type)

    # Update the job status and save the output
    if event_type == 'job.started':
        job.status = 'running'

    if event_type == 'job.suspended':
        # Create a new inbox message with the job trigger output
        inbox_content = await _retrieve_message_content(job_id=agent_job_id)
        inbox_message = models.InboxMessage(
            id=str(uuid.uuid4()),
            job_id=job.id,
            content=inbox_content,
            status="pending",
            received_at=datetime.now()
        )
        db.add(inbox_message)
        job.status = 'suspended'

    if event_type == 'job.completed':
        output_args = agent_job.get('OutputArguments')
        if output_args is not None:
            job.output = output_args.get('final_report', '')
        else:
            job.output = ''
        job.status = 'completed'

    if event_type == 'job.faulted':
        job.status = 'faulted'

    job.completed_at = datetime.now()

    db.commit()

    return {"status": "success", "message": "Job completion processed successfully"}

# API endpoint to get all inbox messages
@app.get("/api/inbox", response_model=List[schemas.InboxMessageResponse])
async def get_inbox_api(
    status: Optional[str] = None,
    skip: int = 0,
    limit: int = 100,
    db: Session = Depends(get_db)
):
    query = db.query(models.InboxMessage)

    if status:
        query = query.filter(models.InboxMessage.status == status)

    messages = query.order_by(models.InboxMessage.received_at.desc()).offset(skip).limit(limit).all()
    return messages

# API endpoint to approve a message
@app.post("/api/approve-message/{message_id}", response_model=schemas.MessageApprovalResponse)
async def approve_message_api(
    message_id: str,
    approval_data: schemas.MessageApprovalRequest,
    background_tasks: BackgroundTasks,
    db: Session = Depends(get_db)
):
    # Find the message
    message = db.query(models.InboxMessage).filter(models.InboxMessage.id == message_id).first()
    if not message:
        raise HTTPException(status_code=404, detail=f"Message with id {message_id} not found")

    # Update message status
    message.status = "approved"
    message.approved_at = datetime.now()
    message.human_feedback = approval_data.human_feedback
    db.commit()

    # Find the associated job
    job = db.query(models.Job).filter(models.Job.id == message.job_id).first()
    if job and job.agent_job_id:
        # Resume the job in the background
        background_tasks.add_task(
            resume_orchestrator_job,
            job_id=job.id,
            agent_job_id=job.agent_job_id,
            human_feedback=approval_data.human_feedback,
            db=db
        )
        return {"status": "success", "message": "Message approved and job resumed with human feedback"}

    return {"status": "success", "message": "Message approved with human feedback"}

# Background task to resume an Orchestrator job
async def resume_orchestrator_job(job_id: str, agent_job_id: str, human_feedback: str, db: Session):
    try:
        payload = True if human_feedback.lower() == "true" else human_feedback

        await sdk.jobs.resume_async(job_id=agent_job_id, payload=payload)

        # Update job status
        job = db.query(models.Job).filter(models.Job.id == job_id).first()
        if job:
            job.status = "resumed"
            job.human_feedback = human_feedback
            db.commit()

    except httpx.HTTPError as e:
        # If the resume request fails, log the error
        job = db.query(models.Job).filter(models.Job.id == job_id).first()
        if job:
            job.error_message = f"Failed to resume job: {str(e)}"
            db.commit()

# API endpoint to get job status
@app.get("/api/job/{job_id}", response_model=schemas.JobDetailResponse)
async def get_job_status_api(job_id: str, db: Session = Depends(get_db)):
    job = db.query(models.Job).filter(models.Job.id == job_id).first()
    if not job:
        raise HTTPException(status_code=404, detail=f"Job with id {job_id} not found")

    # Get associated inbox messages
    messages = db.query(models.InboxMessage).filter(models.InboxMessage.job_id == job_id).all()

    return {
        "id": job.id,
        "topic": job.topic,
        "status": job.status,
        "agent_job_id": job.agent_job_id,
        "created_at": job.created_at,
        "completed_at": job.completed_at,
        "error_message": job.error_message,
        "human_feedback": job.human_feedback,
        "output": job.output,
        "messages": messages
    }

# API endpoint to get all jobs
@app.get("/api/jobs", response_model=List[schemas.JobBasicResponse])
async def get_jobs_api(
    status: Optional[str] = None,
    skip: int = 0,
    limit: int = 100,
    db: Session = Depends(get_db)
):
    query = db.query(models.Job)

    if status:
        query = query.filter(models.Job.status == status)

    jobs = query.order_by(models.Job.created_at.desc()).offset(skip).limit(limit).all()
    return jobs

async def _retrieve_message_content(
    job_id: str,
    folder_key: Optional[str] = None,
    folder_path: Optional[str] = None,
) -> str:
    spec = sdk.jobs._retrieve_inbox_id_spec(
        job_id=job_id,
        folder_key=folder_key,
        folder_path=folder_path,
    )
    response = await sdk.jobs.request_async(
        spec.method,
        url=spec.endpoint,
        params={
            "$filter": f"JobId eq {job_id}",
            "$top": 1
        },
        headers=spec.headers,
    )

    response = response.json()
    print(response)
    if len(response["value"]) > 0:
        return response["value"][0]["TriggerConfiguration"]
    else:
        raise Exception("No inbox found")

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



import os
from typing import List, Optional

import httpx
from langchain_community.callbacks.manager import openai_callback_var
from langchain_openai.embeddings import AzureOpenAIEmbeddings, OpenAIEmbeddings
from pydantic import Field

from uipath_langchain._utils._request_mixin import UiPathRequestMixin
from uipath_langchain._utils._settings import UiPathEndpoints


class UiPathAzureOpenAIEmbeddings(UiPathRequestMixin, AzureOpenAIEmbeddings):
    """Custom Embeddings connector for LangChain integration with UiPath, with minimal changes compared to AzureOpenAIEmbeddings."""

    model_name: Optional[str] = Field(
        default_factory=lambda: os.getenv(
            "UIPATH_MODEL_NAME", "text-embedding-3-large"
        ),
        alias="model",
    )

    def __init__(self, **kwargs):
        super().__init__(
            http_client=httpx.Client(
                event_hooks={
                    "request": [self._log_request_duration],
                    "response": [self._log_response_duration],
                }
            ),
            http_async_client=httpx.AsyncClient(
                event_hooks={
                    "request": [self._alog_request_duration],
                    "response": [self._alog_response_duration],
                }
            ),
            **kwargs,
        )
        self.client._client._prepare_url = self._prepare_url
        self.client._client._build_headers = self._build_headers
        self.async_client._client._prepare_url = self._prepare_url
        self.async_client._client._build_headers = self._build_headers

    @property
    def endpoint(self) -> str:
        return UiPathEndpoints.EMBEDDING_ENDPOINT.value.format(
            model=self.model_name, api_version=self.openai_api_version
        )


class UiPathOpenAIEmbeddings(UiPathRequestMixin, OpenAIEmbeddings):
    """Custom Embeddings connector for LangChain integration with UiPath, with full control over the embedding call."""

    model_name: Optional[str] = Field(
        default_factory=lambda: os.getenv(
            "UIPATH_MODEL_NAME", "text-embedding-3-large"
        ),
        alias="model",
    )

    def embed_documents(
        self, texts: List[str], chunk_size: Optional[int] = None
    ) -> List[List[float]]:
        """Embed a list of documents using the UiPath."""
        embeddings = []
        total_tokens = 0
        # Process in chunks if specified
        chunk_size_ = chunk_size or self.chunk_size or len(texts)
        for i in range(0, len(texts), chunk_size_):
            chunk = texts[i : i + chunk_size_]
            payload = {"input": chunk}
            response = self._call(self.url, payload, self.auth_headers)
            chunk_embeddings = [r["embedding"] for r in response["data"]]
            total_tokens += response["usage"]["prompt_tokens"]
            embeddings.extend(chunk_embeddings)
        if contextvar := openai_callback_var.get():
            contextvar.prompt_tokens += total_tokens
            contextvar.total_tokens += total_tokens
            contextvar.successful_requests += 1
        return embeddings

    async def aembed_documents(
        self,
        texts: List[str],
        chunk_size: Optional[int] = None,
    ) -> List[List[float]]:
        """Async version of embed_documents."""
        embeddings = []
        total_tokens = 0
        # Process in chunks if specified
        chunk_size_ = chunk_size or self.chunk_size or len(texts)
        for i in range(0, len(texts), chunk_size_):
            chunk = texts[i : i + chunk_size_]
            payload = {"input": chunk}
            response = await self._acall(self.url, payload, self.auth_headers)
            chunk_embeddings = [r["embedding"] for r in response["data"]]
            total_tokens += response["usage"]["prompt_tokens"]
            embeddings.extend(chunk_embeddings)
        if contextvar := openai_callback_var.get():
            contextvar.prompt_tokens += total_tokens
            contextvar.total_tokens += total_tokens
            contextvar.successful_requests += 1
        return embeddings

    @property
    def endpoint(self) -> str:
        return UiPathEndpoints.EMBEDDING_ENDPOINT.value.format(
            model=self.model_name, api_version=self.openai_api_version
        )



import json
from typing import Any, Dict, List, Literal, Optional, Union

from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages import AIMessage, BaseMessage
from langchain_core.messages.ai import UsageMetadata
from langchain_core.outputs import ChatGeneration, ChatResult
from langchain_core.runnables import Runnable
from langchain_openai.chat_models import AzureChatOpenAI
from pydantic import BaseModel

from uipath_langchain._utils._request_mixin import UiPathRequestMixin
from uipath_langchain._utils._settings import UiPathEndpoints


class UiPathAzureChatOpenAI(UiPathRequestMixin, AzureChatOpenAI):
    """Custom LLM connector for LangChain integration with UiPath."""

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if "tools" in kwargs and not kwargs["tools"]:
            del kwargs["tools"]
        payload = self._get_request_payload(messages, stop=stop, **kwargs)
        response = self._call(self.url, payload, self.auth_headers)
        return self._create_chat_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if "tools" in kwargs and not kwargs["tools"]:
            del kwargs["tools"]
        payload = self._get_request_payload(messages, stop=stop, **kwargs)
        response = await self._acall(self.url, payload, self.auth_headers)
        return self._create_chat_result(response)

    def with_structured_output(
        self,
        schema: Optional[Any] = None,
        *,
        method: Literal["function_calling", "json_mode", "json_schema"] = "json_schema",
        include_raw: bool = False,
        strict: Optional[bool] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Any]:
        """Model wrapper that returns outputs formatted to match the given schema."""
        schema = (
            schema.model_json_schema()
            if isinstance(schema, type) and issubclass(schema, BaseModel)
            else schema
        )
        return super().with_structured_output(
            schema=schema,
            method=method,
            include_raw=include_raw,
            strict=strict,
            **kwargs,
        )

    @property
    def endpoint(self) -> str:
        return UiPathEndpoints.PASSTHROUGH_COMPLETION_ENDPOINT.value.format(
            model=self.model_name, api_version=self.openai_api_version
        )


class UiPathChat(UiPathRequestMixin, AzureChatOpenAI):
    """Custom LLM connector for LangChain integration with UiPath Normalized."""

    def _create_chat_result(
        self,
        response: Union[Dict[str, Any], BaseModel],
        generation_info: Optional[Dict[Any, Any]] = None,
    ) -> ChatResult:
        if not isinstance(response, dict):
            response = response.model_dump()
        message = response["choices"][0]["message"]
        usage = response["usage"]

        ai_message = AIMessage(
            content=message.get("content", ""),
            usage_metadata=UsageMetadata(
                input_tokens=usage.get("prompt_tokens", 0),
                output_tokens=usage.get("completion_tokens", 0),
                total_tokens=usage.get("total_tokens", 0),
            ),
            additional_kwargs={},
            response_metadata={
                "token_usage": response["usage"],
                "model_name": self.model_name,
                "finish_reason": response["choices"][0].get("finish_reason", None),
                "system_fingerprint": response["id"],
                "created": response["created"],
            },
        )

        if "tool_calls" in message:
            ai_message.tool_calls = [
                {
                    "id": tool["id"],
                    "name": tool["name"],
                    "args": tool["arguments"],
                }
                for tool in message["tool_calls"]
            ]
        generation = ChatGeneration(message=ai_message)
        return ChatResult(generations=[generation])

    def _get_request_payload(
        self,
        input_: LanguageModelInput,
        *,
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[Any, Any]:
        payload = super()._get_request_payload(input_, stop=stop, **kwargs)
        # hacks to make the request work with uipath normalized
        for message in payload["messages"]:
            if message["content"] is None:
                message["content"] = ""
            if "tool_calls" in message:
                for tool_call in message["tool_calls"]:
                    tool_call["name"] = tool_call["function"]["name"]
                    tool_call["arguments"] = json.loads(
                        tool_call["function"]["arguments"]
                    )
            if message["role"] == "tool":
                message["content"] = {
                    "result": message["content"],
                    "call_id": message["tool_call_id"],
                }
        return payload

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """Override the _generate method to implement the chat model logic.

        This can be a call to an API, a call to a local model, or any other
        implementation that generates a response to the input prompt.

        Args:
            messages: the prompt composed of a list of messages.
            stop: a list of strings on which the model should stop generating.
                  If generation stops due to a stop token, the stop token itself
                  SHOULD BE INCLUDED as part of the output. This is not enforced
                  across models right now, but it's a good practice to follow since
                  it makes it much easier to parse the output of the model
                  downstream and understand why generation stopped.
            run_manager: A run manager with callbacks for the LLM.
        """
        if kwargs.get("tools"):
            kwargs["tools"] = [tool["function"] for tool in kwargs["tools"]]
        if "tool_choice" in kwargs and kwargs["tool_choice"]["type"] == "function":
            kwargs["tool_choice"] = {
                "type": "tool",
                "name": kwargs["tool_choice"]["function"]["name"],
            }
        payload = self._get_request_payload(messages, stop=stop, **kwargs)

        response = self._call(self.url, payload, self.auth_headers)
        return self._create_chat_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """Override the _generate method to implement the chat model logic.

        This can be a call to an API, a call to a local model, or any other
        implementation that generates a response to the input prompt.

        Args:
            messages: the prompt composed of a list of messages.
            stop: a list of strings on which the model should stop generating.
                  If generation stops due to a stop token, the stop token itself
                  SHOULD BE INCLUDED as part of the output. This is not enforced
                  across models right now, but it's a good practice to follow since
                  it makes it much easier to parse the output of the model
                  downstream and understand why generation stopped.
            run_manager: A run manager with callbacks for the LLM.
        """
        if kwargs.get("tools"):
            kwargs["tools"] = [tool["function"] for tool in kwargs["tools"]]
        if "tool_choice" in kwargs and kwargs["tool_choice"]["type"] == "function":
            kwargs["tool_choice"] = {
                "type": "tool",
                "name": kwargs["tool_choice"]["function"]["name"],
            }
        payload = self._get_request_payload(messages, stop=stop, **kwargs)

        response = await self._acall(self.url, payload, self.auth_headers)
        return self._create_chat_result(response)

    def with_structured_output(
        self,
        schema: Optional[Any] = None,
        *,
        method: Literal[
            "function_calling", "json_mode", "json_schema"
        ] = "function_calling",
        include_raw: bool = False,
        strict: Optional[bool] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Any]:
        """Model wrapper that returns outputs formatted to match the given schema."""
        if method == "json_schema" and (
            not self.model_name or not self.model_name.startswith("gpt")
        ):
            method = "function_calling"
            if self.logger:
                self.logger.warning(
                    "The json_schema output is not supported for non-GPT models. Using function_calling instead.",
                    extra={
                        "ActionName": self.settings.action_name,
                        "ActionId": self.settings.action_id,
                    }
                    if self.settings
                    else None,
                )
        schema = (
            schema.model_json_schema()
            if isinstance(schema, type) and issubclass(schema, BaseModel)
            else schema
        )
        return super().with_structured_output(
            schema=schema,
            method=method,
            include_raw=include_raw,
            strict=strict,
            **kwargs,
        )

    @property
    def endpoint(self) -> str:
        return UiPathEndpoints.NORMALIZED_COMPLETION_ENDPOINT.value

    @property
    def is_normalized(self) -> bool:
        return True



import asyncio
import json
import os
import uuid
from typing import Any, Dict

from langgraph.graph.state import CompiledStateGraph
from uipath._cli._utils._console import ConsoleLogger
from uipath._cli._utils._parse_ast import generate_bindings_json  # type: ignore
from uipath._cli.middlewares import MiddlewareResult

from ._utils._graph import LangGraphConfig

console = ConsoleLogger()


def resolve_refs(schema, root=None):
    """Recursively resolves $ref references in a JSON schema."""
    if root is None:
        root = schema  # Store the root schema to resolve $refs

    if isinstance(schema, dict):
        if "$ref" in schema:
            ref_path = schema["$ref"].lstrip("#/").split("/")
            ref_schema = root
            for part in ref_path:
                ref_schema = ref_schema.get(part, {})
            return resolve_refs(ref_schema, root)

        return {k: resolve_refs(v, root) for k, v in schema.items()}

    elif isinstance(schema, list):
        return [resolve_refs(item, root) for item in schema]

    return schema


def process_nullable_types(
    schema: Dict[str, Any] | list[Any] | Any,
) -> Dict[str, Any] | list[Any]:
    """Process the schema to handle nullable types by removing anyOf with null and keeping the base type."""
    if isinstance(schema, dict):
        if "anyOf" in schema and len(schema["anyOf"]) == 2:
            types = [t.get("type") for t in schema["anyOf"]]
            if "null" in types:
                non_null_type = next(
                    t for t in schema["anyOf"] if t.get("type") != "null"
                )
                return non_null_type

        return {k: process_nullable_types(v) for k, v in schema.items()}
    elif isinstance(schema, list):
        return [process_nullable_types(item) for item in schema]
    return schema


def generate_schema_from_graph(graph: CompiledStateGraph) -> Dict[str, Any]:
    """Extract input/output schema from a LangGraph graph"""
    schema = {
        "input": {"type": "object", "properties": {}, "required": []},
        "output": {"type": "object", "properties": {}, "required": []},
    }

    if hasattr(graph, "input_schema"):
        if hasattr(graph.input_schema, "model_json_schema"):
            input_schema = graph.input_schema.model_json_schema()
            unpacked_ref_def_properties = resolve_refs(input_schema)

            # Process the schema to handle nullable types
            processed_properties = process_nullable_types(
                unpacked_ref_def_properties.get("properties", {})
            )

            schema["input"]["properties"] = processed_properties
            schema["input"]["required"] = unpacked_ref_def_properties.get(
                "required", []
            )

    if hasattr(graph, "output_schema"):
        if hasattr(graph.output_schema, "model_json_schema"):
            output_schema = graph.output_schema.model_json_schema()
            unpacked_ref_def_properties = resolve_refs(output_schema)

            # Process the schema to handle nullable types
            processed_properties = process_nullable_types(
                unpacked_ref_def_properties.get("properties", {})
            )

            schema["output"]["properties"] = processed_properties
            schema["output"]["required"] = unpacked_ref_def_properties.get(
                "required", []
            )

    return schema


async def langgraph_init_middleware_async(entrypoint: str) -> MiddlewareResult:
    """Middleware to check for langgraph.json and create uipath.json with schemas"""
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        config.load_config()
        entrypoints = []
        all_bindings = {"version": "2.0", "resources": []}
        mermaids = {}

        for graph in config.graphs:
            if entrypoint and graph.name != entrypoint:
                continue

            try:
                loaded_graph = await graph.load_graph()
                state_graph = (
                    loaded_graph.builder
                    if isinstance(loaded_graph, CompiledStateGraph)
                    else loaded_graph
                )
                compiled_graph = state_graph.compile()
                graph_schema = generate_schema_from_graph(compiled_graph)

                mermaids[graph.name] = compiled_graph.get_graph(xray=1).draw_mermaid()

                try:
                    # Make sure the file path exists
                    if os.path.exists(graph.file_path):
                        file_bindings = generate_bindings_json(graph.file_path)

                        # Merge bindings
                        if "resources" in file_bindings:
                            all_bindings["resources"] = file_bindings["resources"]
                except Exception as e:
                    console.warning(
                        f"Warning: Could not generate bindings for {graph.file_path}: {str(e)}"
                    )

                new_entrypoint: dict[str, Any] = {
                    "filePath": graph.name,
                    "uniqueId": str(uuid.uuid4()),
                    "type": "agent",
                    "input": graph_schema["input"],
                    "output": graph_schema["output"],
                }
                entrypoints.append(new_entrypoint)

            except Exception as e:
                console.error(f"Error during graph load: {e}")
                return MiddlewareResult(
                    should_continue=False,
                    should_include_stacktrace=True,
                )
            finally:
                await graph.cleanup()

        if entrypoint and not entrypoints:
            console.error(f"Error: No graph found with name '{entrypoint}'")
            return MiddlewareResult(
                should_continue=False,
            )

        uipath_config = {"entryPoints": entrypoints, "bindings": all_bindings}

        # Save the uipath.json file
        config_path = "uipath.json"
        with open(config_path, "w") as f:
            json.dump(uipath_config, f, indent=2)

        for graph_name, mermaid_content in mermaids.items():
            mermaid_file_path = f"{graph_name}.mermaid"
            try:
                with open(mermaid_file_path, "w") as f:
                    f.write(mermaid_content)
                console.success(f" Created '{mermaid_file_path}' file.")
            except Exception as write_error:
                console.error(
                    f"Error writing mermaid file for '{graph_name}': {str(write_error)}"
                )
                return MiddlewareResult(
                    should_continue=False,
                    should_include_stacktrace=True,
                )
        console.success(f" Created '{config_path}' file.")
        return MiddlewareResult(should_continue=False)

    except Exception as e:
        console.error(f"Error processing langgraph configuration: {str(e)}")
        return MiddlewareResult(
            should_continue=False,
            should_include_stacktrace=True,
        )


def langgraph_init_middleware(entrypoint: str) -> MiddlewareResult:
    """Middleware to check for langgraph.json and create uipath.json with schemas"""
    return asyncio.run(langgraph_init_middleware_async(entrypoint))



import json
import logging
import uuid
from dataclasses import asdict, dataclass
from functools import cached_property
from typing import Any, Dict, Optional, Union, cast

from langgraph.types import Interrupt, StateSnapshot
from uipath import UiPath
from uipath._cli._runtime._contracts import (
    UiPathApiTrigger,
    UiPathErrorCategory,
    UiPathResumeTrigger,
    UiPathResumeTriggerType,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)
from uipath.models import CreateAction, InvokeProcess, WaitAction, WaitJob
from uipath.models.actions import Action

from ._context import LangGraphRuntimeContext
from ._escalation import Escalation
from ._exception import LangGraphRuntimeError

logger = logging.getLogger(__name__)


@dataclass
class InterruptInfo:
    """Contains all information about an interrupt."""

    value: Any

    @property
    def type(self) -> Optional[UiPathResumeTriggerType]:
        """Returns the type of the interrupt value."""
        if isinstance(self.value, CreateAction):
            return UiPathResumeTriggerType.ACTION
        if isinstance(self.value, WaitAction):
            return UiPathResumeTriggerType.ACTION
        if isinstance(self.value, InvokeProcess):
            return UiPathResumeTriggerType.JOB
        if isinstance(self.value, WaitJob):
            return UiPathResumeTriggerType.JOB
        return None

    @property
    def identifier(self) -> Optional[str]:
        """Returns the identifier based on the type."""
        if isinstance(self.value, Action):
            return str(self.value.key)
        return None

    def serialize(self) -> str:
        """
        Converts the interrupt value to a JSON string if possible,
        falls back to string representation if not.
        """
        try:
            if hasattr(self.value, "dict"):
                data = self.value.dict()
            elif hasattr(self.value, "to_dict"):
                data = self.value.to_dict()
            elif hasattr(self.value, "__dataclass_fields__"):
                data = asdict(self.value)
            else:
                data = dict(self.value)

            return json.dumps(data, default=str)
        except (TypeError, ValueError, json.JSONDecodeError):
            return str(self.value)

    @cached_property
    def resume_trigger(self) -> UiPathResumeTrigger:
        """Creates the resume trigger based on interrupt type."""
        if self.type is None:
            return UiPathResumeTrigger(
                api_resume=UiPathApiTrigger(
                    inbox_id=str(uuid.uuid4()), request=self.serialize()
                )
            )
        else:
            return UiPathResumeTrigger(itemKey=self.identifier, triggerType=self.type)


class LangGraphOutputProcessor:
    """
    Contains and manages the complete output information from graph execution.
    Handles serialization, interrupt data, and file output.
    """

    def __init__(self, context: LangGraphRuntimeContext):
        """
        Initialize the LangGraphOutputProcessor.

        Args:
            context: The runtime context for the graph execution.
        """
        self.context = context
        self._interrupt_info: Optional[InterruptInfo] = None
        self._resume_trigger: Optional[UiPathResumeTrigger] = None

        # Process interrupt information during initialization
        state = cast(StateSnapshot, self.context.state)
        if not state or not hasattr(state, "next") or not state.next:
            return

        for task in state.tasks:
            if hasattr(task, "interrupts") and task.interrupts:
                for interrupt in task.interrupts:
                    if isinstance(interrupt, Interrupt):
                        self._interrupt_info = InterruptInfo(interrupt.value)
                        self._resume_trigger = self._interrupt_info.resume_trigger
                        return

    @property
    def status(self) -> UiPathRuntimeStatus:
        """Determines the execution status based on state."""
        return (
            UiPathRuntimeStatus.SUSPENDED
            if self._interrupt_info
            else UiPathRuntimeStatus.SUCCESSFUL
        )

    @property
    def interrupt_value(self) -> Union[Action, InvokeProcess, Any]:
        """Returns the actual value of the interrupt, with its specific type."""
        if self.interrupt_info is None:
            return None
        return self.interrupt_info.value

    @property
    def interrupt_info(self) -> Optional[InterruptInfo]:
        """Gets interrupt information if available."""
        return self._interrupt_info

    @property
    def resume_trigger(self) -> Optional[UiPathResumeTrigger]:
        """Gets resume trigger if interrupted."""
        return self._resume_trigger

    @cached_property
    def serialized_output(self) -> Dict[str, Any]:
        """Serializes the graph execution result."""
        try:
            if self.context.output is None:
                return {}

            return self._serialize_object(self.context.output)

        except Exception as e:
            raise LangGraphRuntimeError(
                "OUTPUT_SERIALIZATION_FAILED",
                "Failed to serialize graph output",
                f"Error serializing output data: {str(e)}",
                UiPathErrorCategory.SYSTEM,
            ) from e

    def _serialize_object(self, obj):
        """Recursively serializes an object and all its nested components."""
        # Handle Pydantic models
        if hasattr(obj, "dict"):
            return self._serialize_object(obj.dict())
        elif hasattr(obj, "model_dump"):
            return self._serialize_object(obj.model_dump(by_alias=True))
        elif hasattr(obj, "to_dict"):
            return self._serialize_object(obj.to_dict())
        # Handle dictionaries
        elif isinstance(obj, dict):
            return {k: self._serialize_object(v) for k, v in obj.items()}
        # Handle lists
        elif isinstance(obj, list):
            return [self._serialize_object(item) for item in obj]
        # Handle other iterable objects (convert to dict first)
        elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes)):
            try:
                return self._serialize_object(dict(obj))
            except (TypeError, ValueError):
                return obj
        # Return primitive types as is
        else:
            return obj

    async def process(self) -> UiPathRuntimeResult:
        """
        Process the output and prepare the final execution result.

        Returns:
            UiPathRuntimeResult: The processed execution result.

        Raises:
            LangGraphRuntimeError: If processing fails.
        """
        try:
            await self._save_resume_trigger()

            return UiPathRuntimeResult(
                output=self.serialized_output,
                status=self.status,
                resume=self.resume_trigger if self.resume_trigger else None,
            )

        except LangGraphRuntimeError:
            raise
        except Exception as e:
            raise LangGraphRuntimeError(
                "OUTPUT_PROCESSING_FAILED",
                "Failed to process execution output",
                f"Unexpected error during output processing: {str(e)}",
                UiPathErrorCategory.SYSTEM,
            ) from e

    async def _save_resume_trigger(self) -> None:
        """
        Stores the resume trigger in the SQLite database if available.

        Raises:
            LangGraphRuntimeError: If database operations fail.
        """
        if not self.resume_trigger or not self.context.memory:
            return

        try:
            await self.context.memory.setup()
            async with (
                self.context.memory.lock,
                self.context.memory.conn.cursor() as cur,
            ):
                try:
                    await cur.execute(f"""
                        CREATE TABLE IF NOT EXISTS {self.context.resume_triggers_table} (
                            id INTEGER PRIMARY KEY AUTOINCREMENT,
                            type TEXT NOT NULL,
                            key TEXT,
                            folder_key TEXT,
                            folder_path TEXT,
                            payload TEXT,
                            timestamp DATETIME DEFAULT (strftime('%Y-%m-%d %H:%M:%S', 'now', 'utc'))
                        )
                    """)
                except Exception as e:
                    raise LangGraphRuntimeError(
                        "DB_TABLE_CREATION_FAILED",
                        "Failed to create resume triggers table",
                        f"Database error while creating table: {str(e)}",
                        UiPathErrorCategory.SYSTEM,
                    ) from e

                try:
                    default_escalation = Escalation()
                    if default_escalation.enabled and isinstance(
                        self.interrupt_value, str
                    ):
                        action = await default_escalation.create(self.interrupt_value)
                        if action:
                            self._resume_trigger = UiPathResumeTrigger(
                                trigger_type=UiPathResumeTriggerType.ACTION,
                                item_key=action.key,
                            )
                    if isinstance(self.interrupt_info, InterruptInfo):
                        uipath_sdk = UiPath()
                        if self.interrupt_info.type is UiPathResumeTriggerType.JOB:
                            if isinstance(self.interrupt_value, InvokeProcess):
                                job = await uipath_sdk.processes.invoke_async(
                                    name=self.interrupt_value.name,
                                    input_arguments=self.interrupt_value.input_arguments,
                                )
                                if job:
                                    self._resume_trigger = UiPathResumeTrigger(
                                        trigger_type=UiPathResumeTriggerType.JOB,
                                        item_key=job.key,
                                    )
                            elif isinstance(self.interrupt_value, WaitJob):
                                self._resume_trigger = UiPathResumeTrigger(
                                    triggerType=UiPathResumeTriggerType.JOB,
                                    itemKey=self.interrupt_value.job.key,
                                )
                        elif self.interrupt_info.type is UiPathResumeTriggerType.ACTION:
                            if isinstance(self.interrupt_value, CreateAction):
                                action = uipath_sdk.actions.create(
                                    title=self.interrupt_value.title,
                                    app_name=self.interrupt_value.app_name
                                    if self.interrupt_value.app_name
                                    else "",
                                    app_folder_path=self.interrupt_value.app_folder_path
                                    if self.interrupt_value.app_folder_path
                                    else "",
                                    app_folder_key=self.interrupt_value.app_folder_key
                                    if self.interrupt_value.app_folder_key
                                    else "",
                                    app_key=self.interrupt_value.app_key
                                    if self.interrupt_value.app_key
                                    else "",
                                    app_version=self.interrupt_value.app_version
                                    if self.interrupt_value.app_version
                                    else 1,
                                    assignee=self.interrupt_value.assignee
                                    if self.interrupt_value.assignee
                                    else "",
                                    data=self.interrupt_value.data,
                                )
                                if action:
                                    self._resume_trigger = UiPathResumeTrigger(
                                        trigger_type=UiPathResumeTriggerType.ACTION,
                                        item_key=action.key,
                                        payload=self.interrupt_value.model_dump_json(),
                                        folder_path=self.interrupt_value.app_folder_path
                                        if self.interrupt_value.app_folder_path
                                        else None,
                                        folder_key=self.interrupt_value.app_folder_key
                                        if self.interrupt_value.app_folder_key
                                        else None,
                                    )
                            elif isinstance(self.interrupt_value, WaitAction):
                                self._resume_trigger = UiPathResumeTrigger(
                                    triggerType=UiPathResumeTriggerType.ACTION,
                                    itemKey=self.interrupt_value.action.key,
                                    payload=self.interrupt_value.model_dump_json(),
                                    folder_path=self.interrupt_value.app_folder_path
                                    if self.interrupt_value.app_folder_path
                                    else None,
                                    folder_key=self.interrupt_value.app_folder_key
                                    if self.interrupt_value.app_folder_key
                                    else None,
                                )

                except Exception as e:
                    raise LangGraphRuntimeError(
                        "ESCALATION_CREATION_FAILED",
                        "Failed to create escalation action",
                        f"Error while creating escalation action: {str(e)}",
                        UiPathErrorCategory.SYSTEM,
                    ) from e

                if (
                    self.resume_trigger.trigger_type.value
                    == UiPathResumeTriggerType.API.value
                    and self.resume_trigger.api_resume
                ):
                    trigger_key = self.resume_trigger.api_resume.inbox_id
                    trigger_type = self.resume_trigger.trigger_type.value
                else:
                    trigger_key = self.resume_trigger.item_key
                    trigger_type = self.resume_trigger.trigger_type.value

                try:
                    logger.debug(f"ResumeTrigger: {trigger_type} {trigger_key}")
                    await cur.execute(
                        f"INSERT INTO {self.context.resume_triggers_table} (type, key, payload, folder_path, folder_key) VALUES (?, ?, ?, ?, ?)",
                        (
                            trigger_type,
                            trigger_key,
                            self.resume_trigger.payload,
                            self.resume_trigger.folder_path,
                            self.resume_trigger.folder_key,
                        ),
                    )
                    await self.context.memory.conn.commit()
                except Exception as e:
                    raise LangGraphRuntimeError(
                        "DB_INSERT_FAILED",
                        "Failed to save resume trigger",
                        f"Database error while saving resume trigger: {str(e)}",
                        UiPathErrorCategory.SYSTEM,
                    ) from e
        except LangGraphRuntimeError:
            raise
        except Exception as e:
            raise LangGraphRuntimeError(
                "RESUME_TRIGGER_SAVE_FAILED",
                "Failed to save resume trigger",
                f"Unexpected error while saving resume trigger: {str(e)}",
                UiPathErrorCategory.SYSTEM,
            ) from e



from typing import Any, Dict, List, Literal, Optional

RUN_TYPE_T = Literal[
    "tool", "chain", "llm", "retriever", "embedding", "prompt", "parser"
]


class CustomTraceEvents:
    UIPATH_TRACE_FUNCTION_CALL = "__uipath_trace_function_call"


class FunctionCallEventData:
    def __init__(
        self,
        function_name: str,
        event_type: str,
        inputs: Dict[str, Any],
        call_uuid: str,
        output: Any,
        error: Optional[str] = None,
        run_type: Optional[RUN_TYPE_T] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.function_name = function_name
        self.event_type = event_type
        self.inputs = inputs
        self.call_uuid = call_uuid
        self.output = output
        self.error = error
        self.run_type = run_type or "chain"
        self.tags = tags
        self.metadata = metadata



import datetime
import logging
from zoneinfo import ZoneInfo


class IgnoreSpecificUrl(logging.Filter):
    def __init__(self, url_to_ignore):
        super().__init__()
        self.url_to_ignore = url_to_ignore

    def filter(self, record):
        try:
            if record.msg == 'HTTP Request: %s %s "%s %d %s"':
                # Ignore the log if the URL matches the one we want to ignore
                method = record.args[0]
                url = record.args[1]

                if method == "POST" and url.path.endswith(self.url_to_ignore):
                    # Check if the URL contains the specific path we want to ignore
                    return True
                return False

        except Exception:
            return False


def _setup_tracer_httpx_logging(url: str):
    # Create a custom logger for httpx
    # Add the custom filter to the root logger
    logging.getLogger("httpx").addFilter(IgnoreSpecificUrl(url))


def _simple_serialize_defaults(obj):
    if hasattr(obj, "model_dump"):
        return obj.model_dump(exclude_none=True, mode="json")
    if hasattr(obj, "dict"):
        return obj.dict()
    if hasattr(obj, "to_dict"):
        return obj.to_dict()

    if isinstance(obj, (set, tuple)):
        if hasattr(obj, "_asdict") and callable(obj._asdict):
            return obj._asdict()
        return list(obj)

    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    if isinstance(obj, (datetime.timezone, ZoneInfo)):
        return obj.tzname(None)

    return str(obj)



from contextlib import asynccontextmanager
from langchain_mcp_adapters.client import MultiServerMCPClient
from langgraph.prebuilt import create_react_agent
from langchain_anthropic import ChatAnthropic
import dotenv
import os

dotenv.load_dotenv()


@asynccontextmanager
async def make_graph():
    async with MultiServerMCPClient() as client:
        await client.connect_to_server_via_sse(
            server_name="hello-world-server",
            url=os.getenv("UIPATH_MCP_SERVER_URL"),
            headers={
                "Authorization": f"Bearer {os.getenv('UIPATH_ACCESS_TOKEN')}"
                },
            timeout=60,
        )

        tools = client.get_tools()
        print(tools)
        model = ChatAnthropic(model="claude-3-5-sonnet-latest")
        graph = create_react_agent(model, tools=tools)
        yield graph



import functools
import importlib
import inspect
import logging
import sys
import uuid
from typing import Any, Callable, Dict, List, Literal, Optional

from langchain_core.callbacks import dispatch_custom_event
from uipath.tracing import TracingManager

from ._events import CustomTraceEvents, FunctionCallEventData

# Original module and traceable function references
original_langsmith: Any = None
original_traceable: Any = None

logger = logging.getLogger(__name__)


def dispatch_trace_event(
    func_name,
    inputs: Dict[str, Any],
    event_type="call",
    call_uuid=None,
    result=None,
    exception=None,
    run_type: Optional[
        Literal["tool", "chain", "llm", "retriever", "embedding", "prompt", "parser"]
    ] = None,
    tags: Optional[List[str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
):
    """Dispatch trace event to our server."""
    try:
        event_data = FunctionCallEventData(
            function_name=func_name,
            event_type=event_type,
            inputs=inputs,
            call_uuid=call_uuid,
            output=result,
            error=str(exception) if exception else None,
            run_type=run_type,
            tags=tags,
            metadata=metadata,
        )

        dispatch_custom_event(CustomTraceEvents.UIPATH_TRACE_FUNCTION_CALL, event_data)
    except Exception as e:
        logger.debug(
            f"Error dispatching trace event: {e}. Function name: {func_name} Event type: {event_type}"
        )


def format_args_for_trace(
    signature: inspect.Signature, *args: Any, **kwargs: Any
) -> Dict[str, Any]:
    try:
        """Return a dictionary of inputs from the function signature."""
        # Create a parameter mapping by partially binding the arguments
        parameter_binding = signature.bind_partial(*args, **kwargs)

        # Fill in default values for any unspecified parameters
        parameter_binding.apply_defaults()

        # Extract the input parameters, skipping special Python parameters
        result = {}
        for name, value in parameter_binding.arguments.items():
            # Skip class and instance references
            if name in ("self", "cls"):
                continue

            # Handle **kwargs parameters specially
            param_info = signature.parameters.get(name)
            if param_info and param_info.kind == inspect.Parameter.VAR_KEYWORD:
                # Flatten nested kwargs directly into the result
                if isinstance(value, dict):
                    result.update(value)
            else:
                # Regular parameter
                result[name] = value

        return result
    except Exception as e:
        logger.warning(
            f"Error formatting arguments for trace: {e}. Using args and kwargs directly."
        )
        return {"args": args, "kwargs": kwargs}


def _create_traced_wrapper(
    func: Callable[..., Any],
    wrapper_func: Optional[Callable[..., Any]] = None,
    func_name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    tags: Optional[List[str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
):
    """Create a traced wrapper based on function type."""
    # Use function name if not provided
    actual_func_name = func_name or func.__name__
    # Function to actually call (the wrapped function or original)
    target_func = wrapper_func or func
    # Ensure we have metadata
    actual_metadata = metadata or {}

    # Define all wrapper functions

    @functools.wraps(target_func)
    async def async_gen_wrapper(*args, **kwargs):
        try:
            call_uuid = str(uuid.uuid4())

            # Get inputs and process them if needed
            inputs = format_args_for_trace(inspect.signature(func), *args, **kwargs)
            if input_processor is not None:
                inputs = input_processor(inputs)

            # Add span_type to metadata if provided
            if span_type:
                actual_metadata["span_type"] = (
                    span_type or "function_call_generator_async"
                )

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "call",
                call_uuid,
                run_type=run_type,  # type: ignore
                tags=tags,
                metadata=actual_metadata,
            )

            outputs = []
            async_gen = target_func(*args, **kwargs)
            async for item in async_gen:
                outputs.append(item)
                yield item

            # Process output if needed
            output_to_record = outputs
            if output_processor is not None:
                output_to_record = output_processor(outputs)

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "completion",
                call_uuid,
                result=output_to_record,
            )
        except Exception as e:
            dispatch_trace_event(
                actual_func_name, inputs, "completion", call_uuid, exception=e
            )
            raise

    @functools.wraps(target_func)
    def gen_wrapper(*args, **kwargs):
        try:
            call_uuid = str(uuid.uuid4())

            # Get inputs and process them if needed
            inputs = format_args_for_trace(inspect.signature(func), *args, **kwargs)
            if input_processor is not None:
                inputs = input_processor(inputs)

            # Add span_type to metadata if provided
            if span_type:
                actual_metadata["span_type"] = (
                    span_type or "function_call_generator_sync"
                )

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "call",
                call_uuid,
                run_type=run_type,  # type: ignore
                tags=tags,
                metadata=actual_metadata,
            )

            outputs = []
            gen = target_func(*args, **kwargs)
            for item in gen:
                outputs.append(item)
                yield item

            # Process output if needed
            output_to_record = outputs
            if output_processor is not None:
                output_to_record = output_processor(outputs)

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "completion",
                call_uuid,
                result=output_to_record,
            )
        except Exception as e:
            dispatch_trace_event(
                actual_func_name, inputs, "completion", call_uuid, exception=e
            )
            raise

    @functools.wraps(target_func)
    async def async_wrapper(*args, **kwargs):
        try:
            call_uuid = str(uuid.uuid4())

            # Get inputs and process them if needed
            inputs = format_args_for_trace(inspect.signature(func), *args, **kwargs)
            if input_processor is not None:
                inputs = input_processor(inputs)

            # Add span_type to metadata if provided
            if span_type:
                actual_metadata["span_type"] = span_type or "function_call_async"

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "call",
                call_uuid,
                run_type=run_type,  # type: ignore
                tags=tags,
                metadata=actual_metadata,
            )

            result = await target_func(*args, **kwargs)

            # Process output if needed
            output_to_record = result
            if output_processor is not None:
                output_to_record = output_processor(result)

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "completion",
                call_uuid,
                result=output_to_record,
            )

            return result
        except Exception as e:
            dispatch_trace_event(
                actual_func_name, inputs, "completion", call_uuid, exception=e
            )
            raise

    @functools.wraps(target_func)
    def sync_wrapper(*args, **kwargs):
        try:
            call_uuid = str(uuid.uuid4())

            # Get inputs and process them if needed
            inputs = format_args_for_trace(inspect.signature(func), *args, **kwargs)
            if input_processor is not None:
                inputs = input_processor(inputs)

            # Add span_type to metadata if provided
            if span_type:
                actual_metadata["span_type"] = span_type or "function_call_sync"

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "call",
                call_uuid,
                run_type=run_type,  # type: ignore
                tags=tags,
                metadata=actual_metadata,
            )

            result = target_func(*args, **kwargs)

            # Process output if needed
            output_to_record = result
            if output_processor is not None:
                output_to_record = output_processor(result)

            dispatch_trace_event(
                actual_func_name,
                inputs,
                "completion",
                call_uuid,
                result=output_to_record,
            )

            return result
        except Exception as e:
            dispatch_trace_event(
                actual_func_name, inputs, "completion", call_uuid, exception=e
            )
            raise

    # Return the appropriate wrapper based on the function type
    if inspect.isasyncgenfunction(target_func):
        return async_gen_wrapper
    elif inspect.isgeneratorfunction(target_func):
        return gen_wrapper
    elif inspect.iscoroutinefunction(target_func):
        return async_wrapper
    else:
        return sync_wrapper


def _create_appropriate_wrapper(
    original_func: Any, wrapped_func: Any, decorator_kwargs: Dict[str, Any]
):
    """Create the appropriate wrapper based on function type."""

    # Get the function name and tags from decorator arguments
    func_name = decorator_kwargs.get("name", original_func.__name__)
    tags = decorator_kwargs.get("tags", None)
    metadata = decorator_kwargs.get("metadata", None)
    run_type = decorator_kwargs.get("run_type", None)

    return _create_traced_wrapper(
        func=original_func,
        wrapper_func=wrapped_func,
        func_name=func_name,
        run_type=run_type,
        tags=tags,
        metadata=metadata,
    )


def _uipath_traced(
    name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
    *args: Any,
    **kwargs: Any,
):
    """Decorator factory that creates traced functions using dispatch_trace_event."""

    def decorator(func):
        return _create_traced_wrapper(
            func=func,
            func_name=name,
            run_type=run_type,
            span_type=span_type,
            input_processor=input_processor,
            output_processor=output_processor,
        )

    return decorator


# Create patched version of traceable
def patched_traceable(*decorator_args, **decorator_kwargs):
    # Handle the case when @traceable is used directly as decorator without arguments
    if (
        len(decorator_args) == 1
        and callable(decorator_args[0])
        and not decorator_kwargs
    ):
        func = decorator_args[0]
        return _create_appropriate_wrapper(func, original_traceable(func), {})

    # Handle the case when @traceable(args) is used with parameters
    original_decorated = original_traceable(*decorator_args, **decorator_kwargs)

    def uipath_trace_decorator(func):
        # Apply the original decorator with its arguments
        wrapped_func = original_decorated(func)
        return _create_appropriate_wrapper(func, wrapped_func, decorator_kwargs)

    return uipath_trace_decorator


# Register the UIPath traced decorator
def register_uipath_tracing():
    """Register the UIPath tracing decorator with TracedDecoratorRegistry."""
    # Reapply to all previously decorated functions
    TracingManager.reapply_traced_decorator(_uipath_traced)


# Apply the patch
def _instrument_traceable_attributes():
    """Apply the patch to langsmith module at import time."""
    global original_langsmith, original_traceable

    # Register our custom tracing decorator
    register_uipath_tracing()

    # Import the original module if not already done
    if original_langsmith is None:
        # Temporarily remove our custom module from sys.modules
        if "langsmith" in sys.modules:
            original_langsmith = sys.modules["langsmith"]
            del sys.modules["langsmith"]

        # Import the original module
        original_langsmith = importlib.import_module("langsmith")

        # Store the original traceable
        original_traceable = original_langsmith.traceable

        # Replace the traceable function with our patched version
        original_langsmith.traceable = patched_traceable

        # Put our modified module back
        sys.modules["langsmith"] = original_langsmith

    return original_langsmith



from ._instrument_traceable import _instrument_traceable_attributes
from .AsyncUiPathTracer import AsyncUiPathTracer
from .UiPathTracer import UiPathTracer

__all__ = [
    "AsyncUiPathTracer",
    "UiPathTracer",
    "_instrument_traceable_attributes",
]



import logging
import os
from os import environ as env
from typing import Generator, Optional

import httpx
import numpy as np
import pytest
from langchain.embeddings import CacheBackedEmbeddings
from langchain.globals import set_llm_cache
from langchain.storage import LocalFileStore
from langchain_community.cache import SQLiteCache
from langchain_community.callbacks import get_openai_callback
from langchain_core.embeddings import Embeddings
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_core.tools import tool
from pydantic import BaseModel, ValidationError

from uipath_langchain._utils._settings import UiPathCachedPathsSettings
from uipath_langchain.chat import (
    UiPathAzureChatOpenAI,
    UiPathChat,
)
from uipath_langchain.embeddings import (
    UiPathAzureOpenAIEmbeddings,
    UiPathOpenAIEmbeddings,
)

test_cache_settings = UiPathCachedPathsSettings(
    CACHED_COMPLETION_DB="tests/llm_cache/tests_uipath_cache.sqlite",
    CACHED_EMBEDDINGS_DIR="tests/llm_cache/cached_embeddings",
)


def get_from_uipath_url():
    try:
        url = os.getenv("UIPATH_URL", "https://cloud.uipath.com/dummyOrg/dummyTennant/")
        if url:
            return "/".join(url.split("/", 3)[:3])
    except Exception:
        return None
    return None


def get_token():
    url_get_token = f"{get_from_uipath_url().rstrip('/')}/identity_/connect/token"

    os.environ["UIPATH_REQUESTING_PRODUCT"] = "uipath-python-sdk"
    os.environ["UIPATH_REQUESTING_FEATURE"] = "langgraph-agent"
    os.environ["UIPATH_TESTS_CACHE_LLMGW"] = "true"

    token_credentials = {
        "client_id": env.get("UIPATH_CLIENT_ID"),
        "client_secret": env.get("UIPATH_CLIENT_SECRET"),
        "grant_type": "client_credentials",
    }

    try:
        with httpx.Client() as client:
            response = client.post(url_get_token, data=token_credentials)
            response.raise_for_status()
            res_json = response.json()
            token = res_json.get("access_token")

            if not token:
                pytest.skip("Authentication token is empty or missing")
    except (httpx.HTTPError, ValueError, KeyError) as e:
        pytest.skip(f"Failed to obtain authentication token: {str(e)}")

    return token


@pytest.fixture(autouse=True)
def setup_test_env():
    env["UIPATH_ACCESS_TOKEN"] = get_token()


@pytest.fixture(scope="module")
def cached_llmgw_calls() -> Generator[Optional[SQLiteCache], None, None]:
    if not os.environ.get("UIPATH_TESTS_CACHE_LLMGW"):
        yield None
    else:
        logging.info("Setting up LLMGW cache")
        db_path = test_cache_settings.cached_completion_db
        os.makedirs(os.path.dirname(db_path), exist_ok=True)
        cache = SQLiteCache(database_path=db_path)
        set_llm_cache(cache)
        yield cache
    set_llm_cache(None)
    return


@pytest.fixture(scope="module")
def cached_embedder() -> Generator[Optional[CacheBackedEmbeddings], None, None]:
    if not os.environ.get("UIPATH_TESTS_CACHE_LLMGW"):
        yield None
    else:
        logging.info("Setting up embeddings cache")
        model = "text-embedding-3-large"
        embedder = CacheBackedEmbeddings.from_bytes_store(
            underlying_embeddings=UiPathOpenAIEmbeddings(model=model),
            document_embedding_cache=LocalFileStore(
                test_cache_settings.cached_embeddings_dir
            ),
            namespace=model,
        )
        yield embedder
    return


def test_cached_call(cached_llmgw_calls: Optional[SQLiteCache]):
    model = UiPathAzureChatOpenAI(cache=cached_llmgw_calls)
    messages = [
        SystemMessage(content="How much is 2 + 2?"),
        HumanMessage(content="Respond with JUST THE ANSWER."),
    ]
    response = model.invoke(messages)
    assert "4" in response.content, f"Expected '4' in response, got: {response.content}"


def test_cached_call_tokens(cached_llmgw_calls: Optional[SQLiteCache]):
    model = UiPathAzureChatOpenAI(cache=cached_llmgw_calls)
    messages = [
        SystemMessage(content="How much is 2 + 2?"),
        HumanMessage(content="Respond with JUST THE ANSWER."),
    ]
    with get_openai_callback() as cb:
        _ = model.invoke(messages)
        total_tokens = cb.total_tokens
    assert total_tokens >= 20, f"Expected more than 20 tokens, got: {total_tokens}"


def test_normalized_cached_call(cached_llmgw_calls: Optional[SQLiteCache]):
    model = UiPathChat(
        model="anthropic.claude-3-5-sonnet-20240620-v1:0", cache=cached_llmgw_calls
    )
    messages = [
        SystemMessage(content="How much is 2 + 2?"),
        HumanMessage(content="Respond with JUST THE ANSWER."),
    ]
    response = model.invoke(messages)
    assert "4" in response.content, f"Expected '4' in response, got: {response.content}"


def test_normalized_cached_call_tokens(cached_llmgw_calls: Optional[SQLiteCache]):
    model = UiPathChat(
        model="anthropic.claude-3-5-sonnet-20240620-v1:0", cache=cached_llmgw_calls
    )
    messages = [
        SystemMessage(content="How much is 2 + 2?"),
        HumanMessage(content="Respond with JUST THE ANSWER."),
    ]
    with get_openai_callback() as cb:
        _ = model.invoke(messages)
        total_tokens = cb.total_tokens
    assert total_tokens >= 20, f"Expected more than 20 tokens, got: {total_tokens}"


def test_tool_call(cached_llmgw_calls: Optional[SQLiteCache]):
    @tool
    def get_first_letter(input):
        """
        Returns the first letter of the input
        """
        return input[0]

    model = UiPathAzureChatOpenAI(cache=cached_llmgw_calls).bind_tools(
        [get_first_letter]
    )
    messages = [
        SystemMessage(content="What is the first letter of the word 'apple'?"),
    ]
    response = model.invoke(messages)
    assert hasattr(response, "tool_calls"), (
        "The response should have a 'tool_calls' attribute"
    )
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "get_first_letter", (
        f"Expected tool call to 'get_first_letter', got: {tool_call['name']}"
    )
    assert tool_call["args"].get("input") == "apple", (
        f"Expected input to be 'apple', got: {tool_call['args'].get('input')}"
    )


def test_structured_output_call(cached_llmgw_calls: Optional[SQLiteCache]):
    class CitySize(BaseModel):
        """City and its size."""

        city: str
        size: int

    model = UiPathAzureChatOpenAI(cache=cached_llmgw_calls).with_structured_output(
        CitySize
    )
    messages = [
        SystemMessage(
            content="What is the capital of France and what is its area in square meters?"
        ),
    ]
    response = model.invoke(messages)
    try:
        city_size = CitySize.model_validate(response)
        assert city_size.city == "Paris", (
            f"Expected city to be 'Paris', got: {city_size.city}"
        )
        assert city_size.size > 0, (
            f"Expected size to be greater than 0, got: {city_size.size}"
        )
    except ValidationError as exc:
        pytest.fail(f"The response was not in the correct format: {exc}")


def test_embedding_call(cached_embedder: Optional[Embeddings]):
    if not cached_embedder:
        cached_embedder = UiPathOpenAIEmbeddings(model="text-embedding-3-large")
    data = [
        "Test input pneumonoultramicroscopicsilicovolcanoconiosis",
        "Another test input",
    ]
    embeds = cached_embedder.embed_documents(data)
    try:
        arr = np.array(embeds)
    except Exception as exc:
        pytest.fail(f"Failed to convert embeddings to numpy array: {exc}")
    assert arr.shape == (2, 3072), f"Expected shape (2, 3072), got: {arr.shape}"


def test_custom_embedding_call(cached_embedder: Optional[Embeddings]):
    if not cached_embedder:
        cached_embedder = UiPathAzureOpenAIEmbeddings(model="text-embedding-3-large")
    data = [
        "Test input pneumonoultramicroscopicsilicovolcanoconiosis",
        "Another test input",
    ]
    embeds = cached_embedder.embed_documents(data)
    try:
        arr = np.array(embeds)
    except Exception as exc:
        pytest.fail(f"Failed to convert embeddings to numpy array: {exc}")
    assert arr.shape == (2, 3072), f"Expected shape (2, 3072), got: {arr.shape}"
